#include "SampleSearch.h"
#include "Solver.h"
#include <zlib.h>
//=================================================================================================
// DIMACS Parser:

#define CHUNK_LIMIT 1048576

class StreamBuffer {
    gzFile  in;
    char    buf[CHUNK_LIMIT];
    int     pos;
    int     size;

    void assureLookahead() {
        if (pos >= size) {
            pos  = 0;
            size = gzread(in, buf, sizeof(buf)); } }

public:
    StreamBuffer(gzFile i) : in(i), pos(0), size(0) {
        assureLookahead(); }

    int  operator *  () { return (pos >= size) ? EOF : buf[pos]; }
    void operator ++ () { pos++; assureLookahead(); }
};

//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

template<class B>
static void skipWhitespace(B& in) {
    while ((*in >= 9 && *in <= 13) || *in == 32)
        ++in; }

template<class B>
static void skipLine(B& in) {
    for (;;){
        if (*in == EOF || *in == '\0') return;
        if (*in == '\n') { ++in; return; }
        ++in; } }

template<class B>
static int parseInt(B& in) {
    int     val = 0;
    bool    neg = false;
    skipWhitespace(in);
    if      (*in == '-') neg = true, ++in;
    else if (*in == '+') ++in;
    if (*in < '0' || *in > '9') reportf("PARSE ERROR! Unexpected char: %c\n", *in), exit(3);
    while (*in >= '0' && *in <= '9')
        val = val*10 + (*in - '0'),
        ++in;
    return neg ? -val : val; }

template<class B>
static void readClause(B& in, Solver& S, vec<Lit>& lits) {
    int     parsed_lit, var;
    lits.clear();
    for (;;){
        parsed_lit = parseInt(in);
        if (parsed_lit == 0) break;
        var = abs(parsed_lit)-1;
        while (var >= S.nVars()) S.newVar();
        lits.push( (parsed_lit > 0) ? Lit(var) : ~Lit(var) );
    }
}

template<class B>
static bool match(B& in, char* str) {
    for (; *str != 0; ++str, ++in)
        if (*str != *in)
            return false;
    return true;
}


template<class B>
static void parse_DIMACS_main(B& in, Solver& S) {
    vec<Lit> lits;
    for (;;){
        skipWhitespace(in);
        if (*in == EOF)
            break;
        else if (*in == 'p'){
            if (match(in, "p cnf")){
                int vars    = parseInt(in);
                int clauses = parseInt(in);
                //reportf("|  Number of variables:  %-12d                                         |\n", vars);
                //reportf("|  Number of clauses:    %-12d                                         |\n", clauses);
            }else{
                reportf("PARSE ERROR! Unexpected char: %c\n", *in), exit(3);
            }
        } else if (*in == 'c' || *in == 'p')
            skipLine(in);
        else
            readClause(in, S, lits),
            S.addClause(lits);
    }
}

// Inserts problem into solver.
//
static void parse_DIMACS(gzFile input_stream, Solver& S) {
    StreamBuffer in(input_stream);
    parse_DIMACS_main(in, S); }


//=================================================================================================

SampleSearch::SampleSearch(GM& gm_, JG& jg_, int p_bound, vector<int>& order_): 
gm(gm_),jg(jg_),order(order_)
{
	helper=IJGPSamplingHelper(&gm,&jg,0,order);
	
	gm.convertToSAT();
}
Double SampleSearch::computePE(int time_limit, int num_samples)
{
	Double weight;
	myRandom random;
	time_t start_time,curr_time;
	start_time=time(NULL);
	Solver solver;
	
	
	
	
	for(int samples=0;samples!=num_samples;samples++)
	{
		gzFile in = gzopen("temp.cnf", "rb");
		parse_DIMACS(in,solver);
		solver.sampler=this;
		curr_time=time(NULL);
		if(num_samples==INVALID_VALUE && (curr_time-start_time) >= time_limit)
		{
			return weight/Double((double)samples);
		}
		// Initialize
		for(int i=0;i<gm.variables.size();i++)
		{
			gm.variables[i]->value()=INVALID_VALUE;
		}
		bool ret_value=solver.solve_sample();
		assert(ret_value);
		//cerr<<"Sample generated\n";
		weight+=getWeight(solver.model);
	}
	return weight/Double((double)num_samples);
}

int SampleSearch::getSample(int& variable, bool& value,vec<char>& assignment)
{
	
	int csp_variable=INVALID_VALUE;
	// Select the earliest unassigned variable

	for(int i=0;i<gm.variables.size();i++)
	{
		int var=order[i];
		
		for(int j=0;j<gm.csp_to_sat_variables[var].size();j++)
		{
			if(toLbool(assignment[gm.csp_to_sat_variables[var][j]])==l_Undef)
			{
				csp_variable=var;
				//cerr<<var<<endl;
				break;
			}
		}
		if(csp_variable==var)
			break;
		
	}
	
	// If all varables are assigned a value return
	if(csp_variable==INVALID_VALUE)
	{
		variable=-1;
		value=false;
		return 1;
	}
	gm.variables[csp_variable]->value()=INVALID_VALUE;
	Double cdf,norm_const;
	vector<Double> func_table(gm.variables[csp_variable]->domain_size());
	//Set values of all variables in the function
	// Assign values to all variables of the function
	
	SF* function=&helper.sampling_functions[csp_variable];
	
	
	for(int i=0;i<function->variables().size();i++)
	{
		if(function->variables()[i]->id()==csp_variable)
			continue;
		function->variables()[i]->value()=INVALID_VALUE;
		int var=function->variables()[i]->id();
		bool val_set=false;
		//function->[i]->resetCurrDomain();
		
		for(int j=0;j<gm.csp_to_sat_variables[var].size();j++)
		{
			if(toLbool(assignment[gm.csp_to_sat_variables[var][j]])==l_True)
			{
				val_set=true;
				gm.variables[var]->value()=j;
				gm.variables[var]->addr_value()=j;
				break;
			}
		}
		assert(val_set);
	}
	assert((int)gm.csp_to_sat_variables[csp_variable].size()==gm.variables[csp_variable]->domain_size());
	int default_value=INVALID_VALUE;
	for(int i=0;i<gm.variables[csp_variable]->domain_size();i++)
	{
		if(toLbool(assignment[gm.csp_to_sat_variables[csp_variable][i]])==l_False)
		{
			continue;
		}
		default_value=i;
		gm.variables[csp_variable]->addr_value()=i;
		int address=Variable::getAddress(function->variables());
		func_table[i]=function->table()[address];
		norm_const+=func_table[i];
	}
	assert(default_value!=INVALID_VALUE);
	for(int i=0;i<gm.variables[csp_variable]->domain_size();i++)
	{
		func_table[i]/=norm_const;
	}
	double rand_val=random.getDouble();
	int sampled_value=INVALID_VALUE;
	for(int i=0;i<gm.variables[csp_variable]->domain_size();i++)
	{
		cdf+=func_table[i];
		if(cdf.value() >=rand_val)
		{
			sampled_value=i;
			//cerr<<func_table[i]<<" ";
			break;
		}
	}
	if(sampled_value==INVALID_VALUE)
	{
		sampled_value=default_value;
	}

	assert(sampled_value!=INVALID_VALUE);
	variable=gm.csp_to_sat_variables[csp_variable][sampled_value];
	value=false;
	
}
Double SampleSearch::getWeight(vec<lbool>& assignment)
{
	Double sample_weight(1.0),actual_weight(1.0);
	Solver solver;
	gzFile in = gzopen("temp.cnf", "rb");
	parse_DIMACS(in,solver);
	vec<Lit> assumptions;
	//cerr<<assignment.size()<<endl;
	// Set value
	for(int v=0;v<gm.variables.size();v++)
	{
		int var=order[v];
		vector<Double> func_table(gm.variables[var]->domain_size());
		SF& function=helper.sampling_functions[var];
		for(int i=0;i<gm.csp_to_sat_variables[var].size();i++)
		{
			gm.variables[var]->addr_value()=i;
			int address=Variable::getAddress(function.variables());
			func_table[i]=function.table()[address];
		}
		bool val_set=false;
		
		for(int i=0;i<gm.csp_to_sat_variables[var].size();i++)
		{
			if(assignment[gm.csp_to_sat_variables[var][i]]==l_True)
			{
				val_set=true;
				gm.variables[var]->addr_value()=i;
				gm.variables[var]->value()=i;
				
				//continue;
			}
			else
			{
				assumptions.push((Lit(gm.csp_to_sat_variables[var][i])));
				//gzFile in = gzopen("temp.cnf", "rb");
				//parse_DIMACS(in,solver);
				if(!solver.solve(assumptions))
				{
					func_table[i]=Double();
				}
				assumptions.pop();
			}
		}
		assert(val_set);
		Double norm_const;
		// Normalize the fun_table
		for(int i=0;i<func_table.size();i++)
		{
			norm_const+=func_table[i];
			
		}
		
		sample_weight*=(func_table[gm.variables[var]->addr_value()]/norm_const);
		
		//cout<<sample_weight<<" "<<func_table[gm.variables[var]->addr_value()]/norm_const<<endl;
		assumptions.push(Lit(gm.csp_to_sat_variables[var][gm.variables[var]->addr_value()]));
	}

	for(int i=0;i<gm.functions.size();i++)
	{
		actual_weight*=gm.functions[i]->getWeight();
	}
	//cout<<"------"<<sample_weight<<" "<<actual_weight<<endl;
	return actual_weight/sample_weight;
}
